// Spell listing

import java.util.ArrayList;
import java.util.Properties;
import java.util.Comparator;
import java.util.Collections;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.awt.*;
import java.io.*;

public class Spells implements ActionListener
{
    public static final String doneCommand = "CLOSE";
    public static final String applyCommand = "APPLY";
    public static final String quitCommand = "QUIT";
    public static final String DATAFILE = "Spells.properties";
    public static final String DATAHEADER = "Dominions Spell & Item list";

    public static final int SC_STRT = 0;
    public static final int SC_CMBT = 1;

    public static void main(String[] args)
    {
        Spells program = new Spells();
        program.run();
    }

    // inner class: Spell
    class Spell
    {
        String name;    // of spell or item
        boolean isItem;
        int itemType;   // if item
        int scope;      // spell scope
        int major;      // major type
        String minor;   // minor types
        Magic[] paths;  // paths & levels needed to cast/make
        Magic school;   // which school & level
        String dshort;  // short description
        String dlong;   // long descrription
        Magic[] cost;   // cost in gems

        public Spell()
        {
            school = new Magic(0,0);
        }

        public void addTo(Properties p, int index)
        {
            p.setProperty("nm"+index, name);
            p.setProperty("Ib"+index, String.valueOf(isItem));
            p.setProperty("it"+index, Integer.toString(itemType));
            p.setProperty("sc"+index, Integer.toString(scope));
            p.setProperty("mj"+index, Integer.toString(major));
            p.setProperty("mn"+index, minor);
            p.setProperty("ds"+index, dshort);
            p.setProperty("dl"+index, dlong);
            if (paths != null)
                for(int i=0; i<paths.length; i++)
                    paths[i].addTo(p, index, "pt", i);
            if (cost != null)
                for(int i=0; i<cost.length; i++)
                    cost[i].addTo(p, index, "cs", i);
            school.addTo(p, index, "sk", 0);
        }

        public boolean loadFrom(Properties p, int index)
        {
            name = p.getProperty("nm"+index);
            if (name == null)
                return false;
            // do the rest of this object
            isItem = Boolean.valueOf((String)p.getProperty("Ib"+index)).booleanValue();
            itemType = Integer.parseInt((String)p.getProperty("it"+index));
            scope = Integer.parseInt((String)p.getProperty("sc"+index));
            major = Integer.parseInt((String)p.getProperty("mj"+index));
            minor = (String)p.getProperty("mn"+index);
            dshort = (String)p.getProperty("ds"+index);
            dlong = (String)p.getProperty("dl"+index);
            cost = new Magic[2];
            paths = new Magic[2];
            for(int i=0; i<2; i++)
            {
                cost[i] = new Magic();
                paths[i] = new Magic();
                cost[i].loadFrom(p, index, "cs", i);
                paths[i].loadFrom(p, index, "pt", i);
            }
            school.loadFrom(p, index, "sk", 0);

            return true;
        }
    }

    // inner class: magic
    class Magic
    {
        int kind;       // -1 indicates fatigue
        int level;      // or amount

        public Magic()
        {
            this(0,0);
        }

        public Magic(int k, int l)
        {
            kind = k;
            level = l;
        }

        public void addTo(Properties p, int index, String hdr, int index2)
        {
            p.setProperty(hdr + "k" + index2 + "_" + index, Integer.toString(kind));
            p.setProperty(hdr + "l" + index2 + "_" + index, Integer.toString(level));
        }

        public void loadFrom(Properties p, int index, String hdr, int index2)
        {
            kind = Integer.parseInt((String)p.getProperty(hdr+"k"+index2+"_"+index));
            level = Integer.parseInt((String)p.getProperty(hdr+"l"+index2+"_"+index));
        }
    }

    class SpellTable implements TableModel
    {
        ArrayList spellList;
        ArrayList myListeners;
        String[] headers;

        public SpellTable(ArrayList theSpells)
        {
            spellList = theSpells;
            myListeners = new ArrayList();
            headers = new String[] { "School", "Paths", "Name", "Costs", "Description" };
        }

        public void addTableModelListener(TableModelListener l)
        {
            myListeners.add(l);
        }

        public Class getColumnClass(int i)
        {
            return headers[i].getClass();
        }

        public int getColumnCount()
        {
            return 5;
        }

        public String getColumnName(int i)
        {
            return headers[i];
        }

        public int getRowCount()
        {
            return spellList.size();
        }

        public Object getValueAt(int row, int col)
        {
            Spell s = (Spell)spellList.get(row);
            return makeDataString(s, col);
        }

        private String makeDataString(Spell s, int col)
        {
            StringBuffer rv = new StringBuffer("");
            int i;

            switch (col)
            {
                case 0:     // school
                    if (s.isItem)     // assume construction
                    {
                        int lvl = s.school.level / 2;
                        rv.append(Spells.itemPower[lvl]).append(" ").append(Spells.itemTypes[s.itemType]);
                    }
                    else if (s.school.kind == Spells.DIVINE)
                    {
                        rv.append(Spells.schoolTypes[s.school.kind]);
                    }
                    else
                    {
                        rv.append(Spells.schoolTypes[s.school.kind]).append(" ").append(s.school.level);
                    }
                    break;

                case 1:     // path
                    for(i=0; i<s.paths.length; i++)
                    {
                        if (s.paths[i].level > 0)
                        {
                            if (i > 0)
                                rv.append(", ");
                            rv.append(Spells.pathTypes[s.paths[i].kind]).append(" ").append(s.paths[i].level);
                        }
                    }
                    break;

                case 2:
                    if (s.name != null)
                        rv.append(s.name);
                    break;

                case 3:     // cost
                    // first check for divine.  If so, no cost.
                    if (s.school.kind == Spells.DIVINE)
                    {
                        rv.append("0");
                        break;
                    }

                    // preliminary:  do we have two gem types?
                    int gemTypes = 0;
                    for(i=0; i<s.cost.length; i++)
                    {
                        if (s.cost[i].level > 0)
                            if (s.cost[i].kind >= 0)
                                gemTypes ++;
                    }

                    for(i=0; i<s.cost.length; i++)
                    {
                        if (s.cost[i].level > 0)
                        {
                            if (i > 0)
                                rv.append(", ");
                            if (s.cost[i].kind < 0)
                                rv.append("f").append(s.cost[i].level).append("-"); // fatigue
                            else
                            {
                                rv.append(s.cost[i].level).append(" ");
                                if (gemTypes >1)        // include path 'air', 'water', etc
                                    rv.append(Spells.pathTypes[s.cost[i].kind]).append(" ");
                                rv.append(Spells.costTypes[s.cost[i].kind]);
                                if (s.cost[i].level > 1)
                                    rv.append("s");
                            }
                        }
                    }
                    if (rv.length() == 0)
                        rv.append("unknown");
                    break;

                case 4:     // short
                    if (s.dshort != null)
                        rv.append(s.dshort);
                    break;
            } // end switch

            return rv.toString();
        }

        public boolean isCellEditable(int row, int column)
        {
            return false;           // table is static for now
        }

        public void removeTableModelListener(TableModelListener l)
        {
            myListeners.remove(l);
        }

        public void setValueAt(Object o, int row, int col) {}    // no-op; RO table

        public Spell getSpellAt(int row)
        {
            return (Spell)spellList.get(row);
        }
    }  // end class SpellTable

    // instance variables
    private ArrayList list;
    private SpellFrame myFrame;
    private volatile Thread thisThread;
    public static String[] pathTypes;
    public static String[] itemPower;
    public static String[] itemTypes;
    public static String[] costTypes;
    public static String[] schoolTypes;
    public static String[] spellTypes;
    public static String[][] spellMinors;
    public static int[] colWidthMax;
    public static int DIVINE = 7;

    static
    {
        pathTypes = new String[] {
            "Air", "Earth", "Fire", "Water", "Nature", "Death", "Astral", "Blood", "Holy", "Unholy" };

        costTypes = new String[] {
            "gem", "gem", "gem", "gem", "gem", "gem", "pearl", "slave", "", "" };

        itemPower = new String[] {
            "Trinket", "Lesser", "Greater", "Powerful", "Unique" };

        itemTypes = new String[] {
            "", "Weapon", "2H Wep", "Shield", "Armour", "Helm", "Boots", "Misc" };

        schoolTypes = new String[] {
            "Conjuration", "Alteration", "Evocation", "Construction", "Enchantment",
            "Thaumaturgy", "Blood", "Divine" };

        spellTypes = new String[] {
            "Personal", "Mobility", "Protection", "Enhance", "Summon", "Harm", "Economic",
            "Misc" };

        spellMinors = new String[][] {
            new String[] { "Mage Enhance", "Melee Enhance", "Enhance", "Defence" },
            null,
            null,
            null,
            null,
            new String[] { "Army", "Undead", "Commander", "Province" },
            new String[] { "Income", "Research", "Gem-related", "Building-related" },
            new String[] { "Control", "Information" }
        };

        colWidthMax = new int[] { 140, 120, 250, 180, 300 };
    }

    public Spells()
    {
        list = new ArrayList();

        // load from disk
        loadSpells();

        // init GUI
        init();
    }

    public void run()
    {
        // show the GUI
        myFrame.setVisible(true);

        // wait on notification
        thisThread = Thread.currentThread();
        while (thisThread != null)
        {
            try
            {
                thisThread.sleep(1000000);
            }
            catch (InterruptedException ie)
            { }
        }

        // dispose of GUI
        myFrame.setVisible(false);
        myFrame.dispose();
        System.exit(0);
    }

    private void init()
    {
        myFrame = new SpellFrame(this, list);
        myFrame.setSize(800,500);
    }

    private void loadSpells()
    {
        // load the entire list from a properties object
        // assume the initial list is empty;  only loaded once
        // during program execution
        Properties p = new Properties();
        int i=0;
        try
        {
            FileInputStream in = new FileInputStream(DATAFILE);

            p.load(in);
            in.close();
        }
        catch(IOException ioe)
        {
            JOptionPane.showMessageDialog(myFrame, "Cannot read spell list",
                "Uh oh!", JOptionPane.WARNING_MESSAGE);
            return;
        }

        // now add items one at a time
        while(true)
        {
            Spell s = new Spell();
            if (!s.loadFrom(p, i))
                break;
            list.add(s);
            i++;
        }
        JOptionPane.showMessageDialog(myFrame, i + " spells loaded");
    }

    private void saveSpells()
    {
        // save the entire list into a properties object
        Properties p = new Properties();
        for(int i=0; i<list.size(); i++)
        {
            Spell s = (Spell)list.get(i);
            s.addTo(p,i);
        }
        try
        {
            FileOutputStream out = new FileOutputStream(DATAFILE);
            p.store(out, DATAHEADER);

            out.close();
        }
        catch (IOException ioe)
        {
            System.out.println(ioe);
            return;
        }
    }

    public void actionPerformed(ActionEvent ae)
    {
        String command = ae.getActionCommand();
        if (command == doneCommand || command == quitCommand)
        {
            if (command == doneCommand)
                // save spell list
                saveSpells();

            // notify ourselves that it's quittin' time
            Thread t = thisThread;
            thisThread = null;
            t.interrupt();
            t = null;
        }
    }

    // end Spells class ----------------------------------------------------

    class SpellFrame extends JFrame implements ItemListener, ActionListener
    {
        private JCheckBox scopeStrt;
        private JCheckBox scopeCmbt;
        private JCheckBox scopeItem;

        private JComboBox majorType;
        private JCheckBox[][] minorChecks;
        private String[] minorTypes;
        private JButton resetButton;
        private JButton allButton;
        private JComboBox[] magicLevels;
        private JButton doneButton;
        private JButton quitButton;

        private JButton newButton;
        private JButton changeButton;
        private JButton removeButton;
        private JTable spellTable;
        private JScrollPane spellPane;

        private CardLayout minorCard;
        private JPanel minorPanel;

        // popup stuff
        private JWindow popupFrame;
        private JLabel  popupLabel;
        private int currentRow = -1;
        private int newRow = -1;
        private Point newPoint = null;
        private long lastMoveTime = 0;

        private ArrayList list;
        private ActionListener ip;

        private boolean doFilter = true;    // to turn off filter

        public SpellFrame(ActionListener parent, ArrayList list)
        {
            super(" Dominions Spell and Item Finder ");
            this.list = list;
            ip = parent;

            // init GUI components
            scopeStrt = new JCheckBox("Strategic");
            scopeCmbt = new JCheckBox("Combat");
            scopeItem = new JCheckBox("Items");

            String[] numbers = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };

            majorType = new JComboBox(Spells.spellTypes);
            resetButton = new JButton("Reset");
            magicLevels = new JComboBox[Spells.pathTypes.length];
            for(int i=0; i<Spells.pathTypes.length; i++)
            {
                magicLevels[i] = new JComboBox(numbers);
                magicLevels[i].setEditable(false);
                magicLevels[i].addItemListener(this);
                magicLevels[i].setMaximumRowCount(10);
            }

            quitButton = new JButton("Quit");
            doneButton = new JButton("Save & Quit");
            resetButton = new JButton("No Paths");
            allButton = new JButton("All Paths");

            newButton = new JButton("New");
            changeButton = new JButton("Change");
            removeButton = new JButton("Remove");

            minorChecks = new JCheckBox[Spells.spellTypes.length][];
            for(int i=0; i<Spells.spellTypes.length; i++)
            {
                if (Spells.spellMinors[i] != null)
                {
                    minorChecks[i] = new JCheckBox[Spells.spellMinors[i].length];
                    for(int j=0; j<Spells.spellMinors[i].length; j++)
                    {
                        minorChecks[i][j] = new JCheckBox(Spells.spellMinors[i][j]);
                        if (i==0)
                            minorChecks[i][j].setSelected(true);
                        minorChecks[i][j].addItemListener(this);
                    }
                }
            }
            spellTable = new JTable(new SpellTable(list));
            spellPane = new JScrollPane(spellTable);
            customizeTable(spellTable);

            // popup stuff
            popupFrame = new JWindow();
            popupLabel = new JLabel();
            popupFrame.getContentPane().add(popupLabel, BorderLayout.NORTH);
            popupFrame.setSize(400,150);
            popupLabel.setMaximumSize(new Dimension(400,500));
            popupLabel.setBackground(Color.blue);
            popupLabel.setForeground(Color.black);
            popupLabel.setBorder(new EtchedBorder(EtchedBorder.RAISED));

            MouseInputAdapter mia = new MouseInputAdapter() {
                public void mouseMoved(MouseEvent me)
                {
                    // algorithm:  every timer tick,  we check the last mouse
                    // move time.   If more than <x> seconds have gone by without
                    // a mouse move,  display the popup for the currently
                    // selected row if it is not already displayed.

                    newRow = spellTable.rowAtPoint(me.getPoint());
                    newPoint = me.getPoint();
                    lastMoveTime = System.currentTimeMillis();
                }

                public void mouseExited(MouseEvent me)
                {
                    if (me.getID() == MouseEvent.MOUSE_EXITED)
                    {
                        // hide popup
                        popupFrame.setVisible(false);
                        // reset row
                        newRow = -1;
                    }
                }

                public void mouseEntered(MouseEvent me)
                {
                    if (me.getID() == MouseEvent.MOUSE_ENTERED)
                    {
                        // just call mouseMoved.
                        mouseMoved(me);
                    }
                }


                };
            spellTable.addMouseListener(mia);
            spellTable.addMouseMotionListener(mia);
            // construct the timer
            java.util.Timer t = new java.util.Timer();
            t.scheduleAtFixedRate(new java.util.TimerTask() {
                public void run()
                {
                    long curTime = System.currentTimeMillis();
                    if (curTime - lastMoveTime > 200
                        && (newRow != currentRow || !popupFrame.isVisible())
                        && newRow != -1)
                    {
                        currentRow = newRow;
                        Point newLoc = new Point(newPoint);
                        Point tableLoc = spellTable.getLocationOnScreen();
                        newLoc.x += tableLoc.x + 5;
                        newLoc.y += tableLoc.y + 5;
                        popupLabel.setText(getToolTipForRow(currentRow));
                        // now that we've set text, get preferred size
                        popupFrame.getContentPane().getLayout().layoutContainer(
                            popupFrame.getContentPane());
                        Dimension d = popupFrame.getContentPane().getLayout()
                            .preferredLayoutSize(popupFrame.getContentPane());
                        d.width = 400;
                        popupFrame.setSize(d);
                        popupFrame.setLocation(newLoc);
                        popupFrame.toFront();
                        popupFrame.setVisible(true);

                    }
                }},  1000, 50 );


            // logic
            majorType.setEditable(false);
            doneButton.addActionListener(ip);
            doneButton.setActionCommand(Spells.doneCommand);
            doneButton.setEnabled(false);
            quitButton.addActionListener(ip);
            quitButton.setActionCommand(Spells.quitCommand);
            scopeStrt.addItemListener(this);
            scopeCmbt.addItemListener(this);
            scopeItem.addItemListener(this);
            majorType.addItemListener(this);
            resetButton.addActionListener(this);
            allButton.addActionListener(this);
            newButton.addActionListener(this);
            changeButton.addActionListener(this);
            removeButton.addActionListener(this);

            // make layout
            makeLayout(getContentPane());

            // initial state
            applyFilter();
        }

        public String getToolTipForRow(int row)
        {
            StringBuffer toolTip = new StringBuffer("<html>");

            Spell s = ((SpellTable)spellTable.getModel()).getSpellAt(row);

            // add name
            toolTip.append("<H2>").append(s.name).append(" - ");
            if (s.isItem)
            {
                // add power, type
                toolTip.append(Spells.itemPower[s.school.level / 2]).append(" ");
                toolTip.append(Spells.itemTypes[s.itemType]);
            }
            else
            {
                // add school
                toolTip.append(Spells.schoolTypes[s.school.kind]);
                if (s.school.kind != Spells.DIVINE)
                    toolTip.append(" level ").append(s.school.level);
            }

            toolTip.append("</h2>");

            if (!s.isItem)
            {
                // add type
                toolTip.append("<H3>").append(Spells.spellTypes[s.major]);
                // add minors
                if (s.minor != null && s.minor.length() > 0)
                {
                    toolTip.append(" - ");
                    for(int i=0; i<s.minor.length(); i++)
                    {
                        if (i>0)
                            toolTip.append(", ");
                        toolTip.append(spellMinors[s.major][Integer.parseInt(s.minor.substring(i,i+1))]);
                    }
                }
            }

            toolTip.append("</H3><P>");

            // path 1
            toolTip.append(Spells.pathTypes[s.paths[0].kind]).append(" ").append(s.paths[0].level);

            // path 2
            if (s.paths.length > 1 && s.paths[1] != null && s.paths[1].level > 0)
                toolTip.append("<BR>").append(Spells.pathTypes[s.paths[1].kind]).append(" ").append(s.paths[1].level);

            // cost 1
            toolTip.append("<P><I>");
            if (s.isItem)
            {
                if (s.cost[0].level != 0)
                {
                    toolTip.append("Requires ").append(s.cost[0].level).append(" ");
                    toolTip.append(pathTypes[s.cost[0].kind]).append(" ");
                    toolTip.append(costTypes[s.cost[0].kind]);
                    if (s.cost[0].level > 1)
                        toolTip.append("s");

                    // second cost - quite common with items
                    if (s.cost[1].level != 0)
                    {
                        toolTip.append(", ").append(s.cost[1].level).append(" ");
                        toolTip.append(pathTypes[s.cost[1].kind]).append(" ");
                        toolTip.append(costTypes[s.cost[1].kind]);
                        if (s.cost[1].level > 1)
                            toolTip.append("s");
                    }
                }
                else
                {
                    // must be unknown - no items are free!
                    toolTip.append("unknown forge cost");
                }
            }
            else
            {
                if (s.cost[0].kind < 0 || s.cost[0].level == 0)
                {
                    toolTip.append("Fatigue Cost: ").append(s.cost[0].level);
                    if (s.cost[0].level != 0)
                        toolTip.append("-");
                }
                else
                    toolTip.append("Ritual<BR>Magic Gem Cost: ").append(s.cost[0].level);

                // cost 2
                if (s.cost.length > 1 && s.cost[1] != null && s.cost[1].level > 0)
                    toolTip.append("<BR>Magic Gem Cost: ").append(s.cost[1].level);
            }


            // break for descrip
            toolTip.append("</I><P>").append(s.dshort);

            if (s.dlong != null)
                toolTip.append("<BR>").append(s.dlong);

            return toolTip.toString();
        }

        private void makeLayout(Container c)
        {
            JPanel p = new JPanel();
            c.add(p);
            p.setLayout(new BorderLayout(10,10));
            p.setBorder(new EmptyBorder(5,5,5,5));

            p.add(makeSelectorPanel(), BorderLayout.NORTH);
            p.add(makeInfoPanel(), BorderLayout.CENTER);
            p.add(makeButtonPanel(), BorderLayout.SOUTH);
        }

        private JPanel makeSelectorPanel()
        {
            JPanel top = new JPanel();
            top.setLayout(new GridLayout(1,4,5,15));

            // grid 1 - scope,  major type
            Box g1 = new Box(BoxLayout.Y_AXIS);
            g1.add(majorType);
            g1.add(Box.createVerticalStrut(25));
            g1.add(scopeStrt);
            g1.add(scopeCmbt);
            g1.add(scopeItem);
            top.add(g1);

            // grid 2 - minor type
            minorCard = new CardLayout();
            minorPanel= new JPanel();
            minorPanel.setLayout(minorCard);
            for(int i=0; i<Spells.spellTypes.length; i++)
            {
                if (minorChecks[i] != null)
                {
                    Box b = new Box(BoxLayout.Y_AXIS);
                    for(int j=0; j<minorChecks[i].length; j++)
                        b.add(minorChecks[i][j]);
                    minorPanel.add(b, Spells.spellTypes[i]);
                }
                else
                {
                    JPanel p = new JPanel();
                    minorPanel.add(p, Spells.spellTypes[i]);
                }
            }
            top.add(minorPanel);

            // grid 3 - spell paths 1
            JPanel p3 = new JPanel();
            p3.setLayout(new BorderLayout());

            JPanel center = new JPanel();
            center.setLayout(new GridLayout(Spells.pathTypes.length / 2, 1, 3,3));
            JPanel east = new JPanel();
            east.setLayout(new GridLayout(Spells.pathTypes.length / 2, 1, 3,3));

            for(int i=0; i<Spells.pathTypes.length / 2; i++)
            {
                center.add(new JLabel(Spells.pathTypes[i] + " up to "));
                Box b = new Box(BoxLayout.X_AXIS);
                b.add(magicLevels[i]);
                b.add(Box.createHorizontalStrut(25));
                east.add(b);
            }
            p3.add(center, BorderLayout.CENTER);
            p3.add(east, BorderLayout.EAST);
            top.add(p3);

            // grid 4 - spell paths 2
            JPanel p4 = new JPanel();
            p4.setLayout(new BorderLayout());

            center = new JPanel();
            center.setLayout(new GridLayout(Spells.pathTypes.length / 2, 1, 3,3));
            east = new JPanel();
            east.setLayout(new GridLayout(Spells.pathTypes.length / 2, 1, 3,3));

            for(int i=Spells.pathTypes.length / 2; i<Spells.pathTypes.length; i++)
            {
                center.add(new JLabel(Spells.pathTypes[i] + " up to "));
                east.add(magicLevels[i]);
            }
            p4.add(center, BorderLayout.CENTER);
            p4.add(east, BorderLayout.EAST);
            top.add(p4);

            return top;
        }

        private JPanel makeInfoPanel()
        {
            JPanel top = new JPanel();
            top.setLayout(new BorderLayout(10,10));

            top.add(spellPane, BorderLayout.CENTER);

            JPanel p = new JPanel();
            p.setLayout(new BorderLayout());

            JPanel b = new JPanel();
            b.setLayout(new GridLayout(3,1,5,5));
            b.add(newButton);
            b.add(changeButton);
            b.add(removeButton);

            p.add(b, BorderLayout.NORTH);
            top.add(p, BorderLayout.EAST);

            return top;
        }

        private JPanel makeButtonPanel()
        {
            JPanel top = new JPanel();

            top.setLayout(new FlowLayout(FlowLayout.CENTER));

            top.add(quitButton);
            top.add(doneButton);
            top.add(resetButton);
            top.add(allButton);

            return top;
        }

        // actionlistener interface
        public void actionPerformed(ActionEvent ae)
        {
            Object source = ae.getSource();

            // try to get currently selected item
            Spell s = null;
            int row = spellTable.getSelectedRow();
            if (row >= 0)
                s = ((SpellTable)spellTable.getModel()).getSpellAt(row);

            if (source == newButton)
            {
                new SpellDialog(this);
            }
            else if (source == changeButton)
            {
                if (s != null)
                    new SpellDialog(this, s);
            }
            else if (source == removeButton)
            {
                if (s != null)
                {
                    list.remove(list.indexOf(s));
                    doneButton.setEnabled(true);
                    applyFilter();
                }
            }
            else if (source == resetButton)
            {
                resetFilter(false);
                applyFilter();
            }
            else if (source == allButton)
            {
                resetFilter(true);
                applyFilter();
            }
        }

        private void resetFilter(boolean which)
        {
            doFilter = false;

            for(int i=0; i<magicLevels.length; i++)
                magicLevels[i].setSelectedIndex((which)?9:0);

            doFilter = true;
        }

        // itemlistener interface
        public void itemStateChanged(ItemEvent ae)
        {
            Object source = ae.getSource();

            // only worry about selected events for combo boxes
            if (source instanceof JComboBox)
                if (ae.getStateChange() != ItemEvent.SELECTED)
                    return;

            // first one to worry about is major type.  Must flip card.
            if (source == majorType)
            {
                String newType = (String)majorType.getSelectedItem();
                minorCard.show(minorPanel, newType);
            }

            // _EVERYTHING_ else that would come here makes us
            // redo the item list
            applyFilter();
        }

        // listselection listener interface

        private void customizeTable(JTable table)
        {
            TableColumn column = null;
            for (int i = 0; i < 5; i++)
            {
                column = table.getColumnModel().getColumn(i);
                column.setMaxWidth(Spells.colWidthMax[i]);
            }
        }

        private void applyFilter()
        {
            if (!doFilter)
                return;

            // get filter criteria - sanity check for empty table
            boolean scStrt = scopeStrt.isSelected();
            boolean scCmbt = scopeCmbt.isSelected();
            boolean scItem = scopeItem.isSelected();

            TableModel t = spellTable.getModel();
            boolean empty = t.getRowCount() == 0;

            if (!(scStrt || scCmbt))
            {
                // early out -- list will be blank
                if (!empty)
                {
                    ArrayList emptyList = new ArrayList();
                    spellTable.setModel(new SpellTable(emptyList));
                    customizeTable(spellTable);
                }
                return;
            }

            // get other filter criteria for more detailed filter
            int major = majorType.getSelectedIndex();
            JCheckBox[] minors = minorChecks[major];

            // get path maxima
            int[] pathMax = new int[Spells.pathTypes.length];
            for(int i=0; i<Spells.pathTypes.length; i++)
                pathMax[i] = magicLevels[i].getSelectedIndex();

            // build new list of spells which pass the filter
            ArrayList newList = new ArrayList();
            for(int i=0; i < list.size(); i++)
            {
                Spell s = (Spell)list.get(i);
                // check type
                if (s.scope == SC_STRT && !scStrt)
                    continue;
                if (s.scope == SC_CMBT && !scCmbt)
                    continue;
                if (s.isItem && !scItem
                    || !s.isItem && scItem)
                    continue;

                // other checks - major, minor type
                if (s.major != major)
                    continue;

                if (minors != null)
                {
                    boolean match = false;
                    // if the spell has no minors,  it will only match if
                    // there are none selected.
                    boolean noMinors = true;
                    for(int z=0; z<minors.length; z++)
                    {
                        if (minors[z].isSelected())
                        {
                            noMinors = false;
                            break;
                        }
                    }

                    // at least one minor must match,  if the major has any
                    if (s.minor == null || s.minor.length() == 0)
                        match = noMinors;
                    else
                    {
                        for(int j=0; j<s.minor.length(); j++)
                        {
                            int q = Integer.parseInt(s.minor.substring(j,j+1));
                            if (minors[q].isSelected())
                            {
                                match = true;
                                break;
                            }
                        }
                    }

                    if (!match)
                        continue;
                }  // end minors

                // check magic paths - none of the spells' paths must
                // exceed the stated maximums
                if (s.paths != null)
                {
                    boolean fail = false;
                    for(int j=0; j<s.paths.length; j++)
                    {
                        if (s.paths[j] != null)
                            if (s.paths[j].level > pathMax[s.paths[j].kind])
                                fail = true;
                    }
                    if (fail)
                        continue;
                }

                newList.add(s);
            }

            // sort the list according to level/shcool
            Collections.sort(newList, new Comparator() {
                public int compare(Object o1, Object o2)
                {
                    Spell s1 = (Spell)o1;
                    Spell s2 = (Spell)o2;
                    int l1 = s1.school.level;
                    int l2 = s2.school.level;

                    // within divine: sort by holy/unholy, then priestly path
                    if (s1.school.kind == Spells.DIVINE
                        && s2.school.kind == Spells.DIVINE)
                        if (s1.paths[0].kind == s2.paths[0].kind)
                            return s1.paths[0].level - s2.paths[0].level;
                        else
                            return s1.paths[0].kind - s2.paths[0].kind;

                    // other vs divine : all spells available immediately - no level.
                    if (s1.school.kind == Spells.DIVINE)
                        l1 = 0;
                    if (s2.school.kind == Spells.DIVINE)
                        l2 = 0;

                    if (l1 == l2)
                        return s1.school.kind - s2.school.kind;
                    else
                        return l1 - l2;
                }
            });

            spellTable.setModel(new SpellTable(newList));
            customizeTable(spellTable);
        }

        public void addChange(Spell s)
        {
            int index = list.indexOf(s);
            if (index >= 0)
                list.set(index, s);
            else
                list.add(s);

            // either way,  better kick the filter and enable the "save & quit" button
            doneButton.setEnabled(true);
            applyFilter();
        }
    }  // end SpellFrame class ----------------------------------------------

    class SpellDialog extends JDialog implements ItemListener, ActionListener, DocumentListener
    {
        private JRadioButton scopeStrt;
        private JRadioButton scopeCmbt;
        private JCheckBox scopeItem;

        private JComboBox majorType;
        private JComboBox itemType;         // for items
        private JCheckBox[][] minorChecks;
        private String[] minorTypes;
        private JComboBox[] pathLevels;     // 2 paths
        private JComboBox[] pathTypes;      // 2 paths
        private JComboBox   magicLevel;     // 1 school
        private JComboBox   magicType;      // 1 school
        private JComboBox[] costTypes;      // which gems/fatigue?: 2 costs
        private JTextField[] cost;          // numeric for cost
        private JTextField shortField;      // short spell desc
        private JTextField name;            // spell/item name.
        private JTextArea  longArea;        // long spell desc.

        private JButton doneButton;
        private JButton applyButton;
        private JButton cancelButton;

        private CardLayout minorCard;
        private JPanel minorPanel;
        private SpellFrame parent;          // I suck!  Hackhackhack

        boolean isEdit = false;             // use add semantic
        Spell spell;

        public SpellDialog(SpellFrame parent)
        {
            this(parent, null);
        }

        public SpellDialog(SpellFrame parent, Spell target)
        {
            super(parent, false);
            this.parent = parent;
            if (target == null)
            {
                isEdit = false;
                spell = new Spell();
            }
            else
            {
                isEdit = true;
                spell = target;
            }
            setTitle(isEdit?" Edit Spell " : " New Spell ");

            // init GUI components
            scopeStrt = new JRadioButton("Strategic");
            scopeCmbt = new JRadioButton("Combat");
            scopeItem = new JCheckBox("Is Item");

            String[] numbers = { "0", "1", "2", "3", "4", "5", "6", "7", "8", "9" };

            majorType = new JComboBox(Spells.spellTypes);
            pathLevels = new JComboBox[2];
            pathLevels[0] = new JComboBox(numbers);
            pathLevels[0].setMaximumRowCount(10);
            pathLevels[1] = new JComboBox(numbers);
            pathLevels[1].setMaximumRowCount(10);
            pathTypes = new JComboBox[2];
            pathTypes[0] = new JComboBox(Spells.pathTypes);
            pathTypes[1] = new JComboBox(Spells.pathTypes);
            magicLevel = new JComboBox(numbers);
            magicLevel.setMaximumRowCount(10);
            magicType = new JComboBox(Spells.schoolTypes);
            costTypes = new JComboBox[2];
            costTypes[0] = new JComboBox(Spells.pathTypes);
            costTypes[0].addItem("Fatigue");
            costTypes[0].setMaximumRowCount(10);
            costTypes[1] = new JComboBox(Spells.pathTypes);
            shortField = new JTextField();
            name = new JTextField();
            longArea = new JTextArea();
            cost = new JTextField[2];
            cost[0] = new JTextField("0");
            cost[1] = new JTextField("0");
            itemType = new JComboBox(Spells.itemTypes);

            doneButton = new JButton("Done");
            applyButton = new JButton("Apply");
            cancelButton = new JButton("Cancel");

            minorChecks = new JCheckBox[Spells.spellTypes.length][];
            for(int i=0; i<Spells.spellTypes.length; i++)
            {
                if (Spells.spellMinors[i] != null)
                {
                    minorChecks[i] = new JCheckBox[Spells.spellMinors[i].length];
                    for(int j=0; j<Spells.spellMinors[i].length; j++)
                        minorChecks[i][j] = new JCheckBox(Spells.spellMinors[i][j]);
                }
            }

            // logic
            majorType.setEditable(false);
            majorType.addItemListener(this);
            doneButton.addActionListener(this);
            applyButton.addActionListener(this);
            cancelButton.addActionListener(this);
            scopeCmbt.addItemListener(this);
            scopeItem.addItemListener(this);
            pathTypes[0].addItemListener(this);
            magicLevel.addItemListener(this);
            pathLevels[0].addItemListener(this);
            name.getDocument().addDocumentListener(this);
            shortField.getDocument().addDocumentListener(this);
            ButtonGroup bg1 = new ButtonGroup();
            bg1.add(scopeStrt);
            bg1.add(scopeCmbt);

            // make layout
            makeLayout(getContentPane());

            // initial state
            updateGUI();
            enableButtons();

            setSize(800,300);
            setVisible(true);
        }

        private void enableButtons()
        {
            boolean enable = false;
            // check path1 level, name, field
            if (pathLevels[0].getSelectedIndex() > 0
                && name.getText().length() > 0
                && shortField.getText().length() > 0)
                enable = true;

            doneButton.setEnabled(enable);
            applyButton.setEnabled(enable);
        }

        private void updateSpell(Spell s)
        {
            s.name = name.getText();
            s.dshort = shortField.getText();
            s.dlong = longArea.getText();

            s.scope = scopeStrt.isSelected()?SC_STRT:SC_CMBT;
            s.isItem = scopeItem.isSelected();
            if (s.isItem)
                s.itemType = itemType.getSelectedIndex();
            s.major = majorType.getSelectedIndex();
            s.minor = "";
            if (minorChecks[s.major] != null)
            {
                for(int i=0; i<minorChecks[s.major].length; i++)
                {
                    if (minorChecks[s.major][i].isSelected())
                        s.minor += Integer.toString(i);
                }
            }
            s.school.kind = magicType.getSelectedIndex();
            s.school.level = magicLevel.getSelectedIndex();
            s.paths = new Magic[2];
            s.paths[0] = new Magic(pathTypes[0].getSelectedIndex(), pathLevels[0].getSelectedIndex());
            s.paths[1] = new Magic(pathTypes[1].getSelectedIndex(), pathLevels[1].getSelectedIndex());
            s.cost = new Magic[2];
            s.cost[0] = new Magic(costTypes[0].getSelectedIndex(), Integer.parseInt(cost[0].getText()));
            s.cost[1] = new Magic(costTypes[1].getSelectedIndex(), Integer.parseInt(cost[1].getText()));
            // tweak for fatigue
            if (s.cost[0].kind == Spells.pathTypes.length)
                s.cost[0].kind = -1;
        }

        private void updateGUI()
        {
            if (spell.name != null) name.setText(spell.name);
                else name.setText("");
            if (spell.dshort != null) shortField.setText(spell.dshort);
                else shortField.setText("");
            if (spell.dlong != null) longArea.setText(spell.dlong);
                else longArea.setText("");

            scopeItem.setSelected(spell.isItem);
            if (spell.scope == Spells.SC_STRT)
                scopeStrt.setSelected(true);
            else
                scopeCmbt.setSelected(true);
            if (spell.isItem)
                itemType.setSelectedIndex(spell.itemType);

            majorType.setSelectedIndex(spell.major);
            clearMinors();
            if (spell.minor != null)
            {
                for(int i=0; i<spell.minor.length(); i++)
                {
                    String mnr = spell.minor.substring(i,i+1);
                    minorChecks[spell.major][Integer.parseInt(mnr)].setSelected(true);
                }
            }
            magicType.setSelectedIndex(spell.school.kind);
            magicLevel.setSelectedIndex(spell.school.level);

            if (spell.paths != null)
            {
                for(int i=0; i<spell.paths.length; i++)
                {
                    if (spell.paths[i] != null)
                    {
                        pathTypes[i].setSelectedIndex(spell.paths[i].kind);
                        pathLevels[i].setSelectedIndex(spell.paths[i].level);
                    }
                }
            }

            if (spell.cost != null)
            {
                if (spell.cost.length > 0)
                {
                    if (spell.cost[0].kind < 0)
                        costTypes[0].setSelectedIndex(Spells.pathTypes.length);
                    else
                        costTypes[0].setSelectedIndex(spell.cost[0].kind);
                    cost[0].setText(Integer.toString(spell.cost[0].level));
                }
                if (spell.cost.length > 1)
                {
                    if (spell.cost[1] != null)
                    {
                        costTypes[1].setSelectedIndex(spell.cost[1].kind);
                        cost[1].setText(Integer.toString(spell.cost[1].level));
                    }
                }
            }
        }

        private void clearMinors()
        {
            int i;

            for(i=0; i<Spells.spellTypes.length; i++)
                if (minorChecks[i] != null)
                {
                    for(int j=0; j<minorChecks[i].length; j++)
                        minorChecks[i][j].setSelected(false);
                }
        }


        private void makeLayout(Container c)
        {
            JPanel p = new JPanel();
            c.add(p);
            p.setLayout(new BorderLayout(10,10));
            p.setBorder(new EmptyBorder(5,5,5,5));

            p.add(makeSelectorPanel(), BorderLayout.NORTH);
            p.add(makeTextPanel(), BorderLayout.CENTER);
            p.add(makeButtonPanel(), BorderLayout.SOUTH);
        }

        private JPanel makeSelectorPanel()
        {
            JPanel top = new JPanel();
            top.setLayout(new GridLayout(1,4,5,15));

            // grid 1 - scope,  major type
            Box g1 = new Box(BoxLayout.Y_AXIS);
            g1.add(majorType);
            g1.add(Box.createVerticalStrut(25));
            g1.add(scopeStrt);
            g1.add(scopeCmbt);
            Box ib = new Box(BoxLayout.X_AXIS);
            ib.add(scopeItem);
            ib.add(Box.createHorizontalStrut(15));
            ib.add(itemType);
            g1.add(ib);
            top.add(g1);

            // grid 2 - minor type
            minorCard = new CardLayout();
            minorPanel= new JPanel();
            minorPanel.setLayout(minorCard);
            for(int i=0; i<Spells.spellTypes.length; i++)
            {
                if (minorChecks[i] != null)
                {
                    Box b = new Box(BoxLayout.Y_AXIS);
                    for(int j=0; j<minorChecks[i].length; j++)
                        b.add(minorChecks[i][j]);
                    minorPanel.add(b, Spells.spellTypes[i]);
                }
                else
                {
                    JPanel p = new JPanel();
                    minorPanel.add(p, Spells.spellTypes[i]);
                }
            }
            top.add(minorPanel);

            // grid 3 - school, path1, path2, cost 1
            JPanel p3 = new JPanel();
            p3.setLayout(new BorderLayout());

            JPanel center = new JPanel();
            center.setLayout(new GridLayout(4, 1, 3,3));
            JPanel east = new JPanel();
            east.setLayout(new GridLayout(4, 1, 3,3));

            center.add(new JLabel("School "));
            center.add(new JLabel("Path 1 "));
            center.add(new JLabel("Path 2 "));
            Box b = new Box(BoxLayout.X_AXIS);
            b.add(new JLabel("Cost: "));
            b.add(cost[0]);
            b.add(Box.createHorizontalStrut(10));
            center.add(b);

            east.add(magicType);
            east.add(pathTypes[0]);
            east.add(pathTypes[1]);
            east.add(costTypes[0]);
            p3.add(center, BorderLayout.CENTER);
            p3.add(east, BorderLayout.EAST);
            top.add(p3);

            // grid 4 - school level, path1 level, path2 level, cost 2
            JPanel p4 = new JPanel();
            p4.setLayout(new BorderLayout());

            center = new JPanel();
            center.setLayout(new GridLayout(4, 1, 3,3));
            east = new JPanel();
            east.setLayout(new GridLayout(4, 1, 3,3));

            center.add(new JLabel(" level "));
            center.add(new JLabel(" power "));
            center.add(new JLabel(" power "));
            b = new Box(BoxLayout.X_AXIS);
            b.add(new JLabel(" and "));
            b.add(cost[1]);
            b.add(Box.createHorizontalStrut(10));
            center.add(b);

            east.add(magicLevel);
            east.add(pathLevels[0]);
            east.add(pathLevels[1]);
            east.add(costTypes[1]);

            p4.add(center, BorderLayout.CENTER);
            p4.add(east, BorderLayout.EAST);
            top.add(p4);

            return top;
        }

        private JPanel makeTextPanel()
        {
            JPanel top = new JPanel();
            top.setLayout(new BorderLayout(5,5));

            JPanel north = new JPanel();
            north.setLayout(new GridLayout(1,2,5,5));
            // name
            Box b1 = new Box(BoxLayout.X_AXIS);
            b1.add(new JLabel("Name: "));
            b1.add(name);
            b1.add(Box.createHorizontalStrut(8));
            north.add(b1);

            // short desc.
            Box b2 = new Box(BoxLayout.X_AXIS);
            b2.add(new JLabel("Short: "));
            b2.add(shortField);
            north.add(b2);

            top.add(north, BorderLayout.NORTH);

            // long desc
            JScrollPane center = new JScrollPane(longArea);
            top.add(center, BorderLayout.CENTER);

            return top;
        }

        private JPanel makeButtonPanel()
        {
            JPanel top = new JPanel();

            top.setLayout(new FlowLayout(FlowLayout.RIGHT));

            top.add(doneButton);
            top.add(applyButton);
            top.add(cancelButton);

            return top;
        }

        // actionlistener interface
        public void actionPerformed(ActionEvent ae)
        {
            Object source = ae.getSource();

            if (source == cancelButton)
            {
                setVisible(false);
                dispose();
            }
            else if (source == doneButton)
            {
                applyChanges();
                setVisible(false);
                dispose();
            }
            else if (source == applyButton)
            {
                applyChanges();
                if (!isEdit)
                {
                    // make a new one,  but keep school
                    Magic m = spell.school;
                    boolean wasItem = spell.isItem;
                    int iType = spell.itemType;
                    spell = new Spell();
                    spell.school = new Magic(m.kind, m.level);

                    // init paths & costs
                    Magic[] newPaths = new Magic[2];
                    Magic[] newCosts = new Magic[2];
                    for(int i=0; i<2; i++)
                    {
                        newPaths[i] = new Magic(0,0);
                        newCosts[i] = new Magic(0,0);
                    }
                    spell.paths = newPaths;
                    spell.cost = newCosts;

                    // leave item unchanged
                    spell.isItem = wasItem;
                    spell.itemType = iType;

                    updateGUI();
                }
            }
        }

        private void applyChanges()
        {
            updateSpell(spell);
            parent.addChange(spell);
        }

        // itemlistener interface
        public void itemStateChanged(ItemEvent ae)
        {
            Object source = ae.getSource();

            // only worry about selected events for combo boxes
            if (source instanceof JComboBox)
                if (ae.getStateChange() != ItemEvent.SELECTED)
                    return;

            // first one to worry about is major type.  Must flip card.
            if (source == majorType)
            {
                String newType = (String)majorType.getSelectedItem();
                minorCard.show(minorPanel, newType);
            }
            else if (source == scopeCmbt)
            {
                // select fatigue, if its not a spell
                if (scopeCmbt.isSelected() && !scopeItem.isSelected())
                    costTypes[0].setSelectedIndex(Spells.pathTypes.length);
            }
            else if (source == scopeItem)
            {
                // set school to construction
                magicType.setSelectedIndex(3);      // ooh, a magic number!

                // set cost to path0
                costTypes[0].setSelectedIndex(pathTypes[0].getSelectedIndex());
            }
            else if (source == pathTypes[0])
            {
                if (scopeCmbt.isSelected() && !scopeItem.isSelected())
                    costTypes[1].setSelectedIndex(pathTypes[0].getSelectedIndex());
                else
                    costTypes[0].setSelectedIndex(pathTypes[0].getSelectedIndex());
            }
            else if (source == pathLevels[0] || source == magicLevel)
            {
                enableButtons();
            }
        }

        // doclistener interface
        public void changedUpdate(DocumentEvent de)
        {
            enableButtons();
        }

        public void insertUpdate(DocumentEvent de) { changedUpdate(de); }
        public void removeUpdate(DocumentEvent de) { changedUpdate(de); }

    } // end SpellDialog class -----------------------------------------------

}


